/***********************************************************************/
/*                                                                     */
/*                           Objective Caml                            */
/*                                                                     */
/*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
/*                                                                     */
/*  Copyright 1996 Institut National de Recherche en Informatique et   */
/*  en Automatique.  All rights reserved.  This file is distributed    */
/*  under the terms of the GNU Library General Public License, with    */
/*  the special exception on linking described in file ../../LICENSE.  */
/*                                                                     */
/***********************************************************************/

/* $Id: text.c,v 1.15 2004/03/24 15:02:06 starynke Exp $ */

/* Patched by Jean Mouric for use with MacCaml: 2009/09/30 */

#include <locale.h>
#include "libgraph.h"
#include <caml/alloc.h>
#include <string.h>

XFontStruct * caml_gr_font = NULL;

static void caml_gr_get_font(char *fontname)
{
  XLockDisplay(caml_gr_display);
  XFontStruct * font = XLoadQueryFont(caml_gr_display, fontname);
  if (font == NULL) caml_gr_fail("cannot find font %s", fontname);
  if (caml_gr_font != NULL) XFreeFont(caml_gr_display, caml_gr_font);
  caml_gr_font = font;
  XSetFont(caml_gr_display, caml_gr_window.gc, caml_gr_font->fid);
  XSetFont(caml_gr_display, caml_gr_bstore.gc, caml_gr_font->fid);
  XUnlockDisplay(caml_gr_display);
}

value caml_gr_set_font(value fontname)
{
  caml_gr_check_open();
  caml_gr_get_font(String_val(fontname));
  return Val_unit;
}

value caml_gr_set_text_size (value sz)
{
  return Val_unit;
}

static void caml_gr_draw_text(char *txt, int len)
{
  if (caml_gr_font == NULL) caml_gr_get_font(DEFAULT_FONT);
  XLockDisplay(caml_gr_display);
  if (caml_gr_remember_modeflag)
    XDrawString(caml_gr_display, caml_gr_bstore.win, caml_gr_bstore.gc,
                caml_gr_x, Bcvt(caml_gr_y) - caml_gr_font->descent + 1, txt, len);
  if (caml_gr_display_modeflag) {
    XDrawString(caml_gr_display, caml_gr_window.win, caml_gr_window.gc,
                caml_gr_x, Wcvt(caml_gr_y) - caml_gr_font->descent + 1, txt, len);
    XFlush(caml_gr_display);
  }
  XUnlockDisplay(caml_gr_display);
  caml_gr_x += XTextWidth(caml_gr_font, txt, len);
}

int text_length(char * str)
{
    size_t lenA = strlen(str);
    setlocale(LC_CTYPE, "fr_FR.UTF-8");
    size_t lenW = mbstowcs( NULL, str, 0);
    return lenW == (size_t)-1 ? (int)lenA : (int)lenW;
}

value caml_gr_text_length(value str)
{
    return Val_int(text_length(String_val(str)));
}

static XFontSet font=NULL;
static int ascent = 0;
static int descent = 0;

static void create_font()
{
    if (font == NULL) {
        setlocale(LC_CTYPE, "fr_FR.UTF-8");
        char **missing_charset_list;
        int missing_charset_count;
        char *def_string;
        font = XCreateFontSet(caml_gr_display, "*-fixed-*-r-*-*-*-*-*-*-*,*", &missing_charset_list, &missing_charset_count, &def_string);
        if (font) {
            XFontStruct **xfonts = NULL;
            char **font_names = NULL;
            XFontSetExtents *font_extents = XExtentsOfFontSet(font);
            int fontcount = XFontsOfFontSet(font, &xfonts, &font_names);
            int i;
            for(i = 0, ascent = 0, descent = 0; i<fontcount; i++){
                if (ascent < (*xfonts)->ascent) ascent = (*xfonts)->ascent;
                if (descent < (*xfonts)->descent) descent = (*xfonts)->descent;
                xfonts++;
            }
        }
    }
}

static void caml_gr_draw_text_utf8(txt, len)
char * txt;
int len;
{
    if (caml_gr_font == NULL) caml_gr_get_font(DEFAULT_FONT);
        XLockDisplay(caml_gr_display);
        create_font();
        if (font) {
            setlocale(LC_CTYPE, "fr_FR.UTF-8");
            Xutf8DrawString(caml_gr_display, caml_gr_window.win, font, caml_gr_window.gc,
                            caml_gr_x, Wcvt(caml_gr_y) - descent + 1, txt, len);
            Xutf8DrawString(caml_gr_display, caml_gr_bstore.win, font, caml_gr_bstore.gc,
                            caml_gr_x, Bcvt(caml_gr_y) - descent + 1, txt, len);
            caml_gr_x += Xutf8TextEscapement(font, txt, len);
        }
        else {
            XDrawString(caml_gr_display, caml_gr_window.win, caml_gr_window.gc,
                        caml_gr_x, Wcvt(caml_gr_y) - caml_gr_font->descent + 1, txt, len);
            XDrawString(caml_gr_display, caml_gr_bstore.win, caml_gr_bstore.gc,
                        caml_gr_x, Bcvt(caml_gr_y) - caml_gr_font->descent + 1, txt, len);
            caml_gr_x += XTextWidth(caml_gr_font, txt, len);
        }
    XFlush(caml_gr_display);
    XUnlockDisplay(caml_gr_display);
}

value caml_gr_draw_char(value chr)
{
  char str[1];
  caml_gr_check_open();
  str[0] = Int_val(chr);
  caml_gr_draw_text(str, 1);
  return Val_unit;
}

value caml_gr_draw_string(value str)
{
  caml_gr_check_open();
  caml_gr_draw_text_utf8(String_val(str), caml_string_length(str));
  return Val_unit;
}

value caml_gr_text_size(value str)
{
  int width;
    int height;
  value res;
  caml_gr_check_open();
  XLockDisplay(caml_gr_display);
  if (caml_gr_font == NULL) caml_gr_get_font(DEFAULT_FONT);
  create_font();
  if (font) {
    width = Xutf8TextEscapement(font, String_val(str), caml_string_length(str));
    height = ascent + descent;
  }
  else {
    width = XTextWidth(caml_gr_font, String_val(str), caml_string_length(str));
    height = caml_gr_font->ascent + caml_gr_font->descent;
  }
  XUnlockDisplay(caml_gr_display);
  res = caml_alloc_small(2, 0);
  Field(res, 0) = Val_int(width);
  Field(res, 1) = Val_int(height);
  return res;
}
